import re
import struct
import base64
import hashlib

from pywss.public import WebSocketProtocolError

MAGIC_STRING = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
REGEX = re.compile(r'GET\s+([^\s]+).*Sec-WebSocket-Key:\s*(.*?)\r\n', re.S)
RESPONSE_TEMPLATE = "HTTP/1.1 101 Switching Protocols\r\n" \
                    "Upgrade: websocket\r\n" \
                    "Connection: Upgrade\r\n" \
                    "Sec-WebSocket-Accept: %s\r\n\r\n"


class WebSocketProtocol:

    @classmethod
    def check_header(cls, headers):
        try:
            path, key = REGEX.search(headers).groups()
            if all((path, key)):
                return path, key
        except:
            pass
        raise WebSocketProtocolError

    @classmethod
    def get_ac_str(cls, data):
        path, value = cls.check_header(str(data, encoding="utf-8"))
        return path, base64.b64encode(hashlib.sha1((value + MAGIC_STRING).encode('utf-8')).digest()).decode('utf-8')

    @classmethod
    def parsing(cls, data):
        path, key = cls.get_ac_str(data)
        return path, bytes(RESPONSE_TEMPLATE % key, encoding='utf-8')

    @classmethod
    def decode_msg(cls, data):
        payload_len = data[1] & 0b1111111
        if payload_len is 0b1111110:
            mask = data[4:8]
            decoded = data[8:]
        elif payload_len is 0b1111111:
            mask = data[10:14]
            decoded = data[14:]
        else:
            mask = data[2:6]
            decoded = data[6:]
        return str(bytearray([decoded[i] ^ mask[i % 4] for i in range(len(decoded))]), encoding='utf-8')

    @classmethod
    def encode_msg(cls, msg_bytes, token=b"\x81"):
        length = len(msg_bytes)
        if length <= 125:
            token += struct.pack("B", length)
        elif length <= 65535:
            token += struct.pack("!BH", 126, length)
        else:
            token += struct.pack("!BQ", 127, length)
        return token + msg_bytes



from queue import Queue

radio_queue = Queue()
RADIO_START = object()
ERROR_FLAG = object()


class PublicConfig:
    DEFAULT_REPLY = "Hello WebSocket"
    ERROR_COUNT_MAX = 10
    RADIO_TIME = 10


class AuthenticationError(Exception):
    """Cookie Auth fail"""


class WebSocketProtocolError(Exception):
    """Protocol Auth fail"""


class OutputTypeError(Exception):
    """Output error"""


class InvalidPath(Exception):
    """path error"""


class MiddlewareError(Exception):
    """middleware error"""



import time


# SnowFlake
class SnowKey:
    def __init__(self, workerId, datacenterId):
        self.workerId = workerId
        self.datacenterId = datacenterId
        self.sequence = 0
        self.twepoch = 1288834974657
        self.workerIdBits = 5
        self.datacenterIdBits = 5
        self.maxWorkerId = -1 ^ (-1 << self.workerIdBits)
        self.maxDatacenterId = -1 ^ (-1 << self.datacenterIdBits)
        self.sequenceBits = 12
        self.workerIdShift = self.sequenceBits
        self.datacenterIdShift = self.sequenceBits + self.workerIdBits
        self.timestampLeftShift = self.sequenceBits + self.workerIdBits + self.datacenterIdBits
        self.sequenceMask = -1 ^ (-1 << self.sequenceBits)
        self.lastTimestamp = -1
        self.get_current_timestamp = lambda: int(time.time() * 1000)

    @classmethod
    def from_node(cls, workerId=0, datacenterId=0):
        return cls(workerId, datacenterId)

    def next_id(self):
        timestamp = self.get_current_timestamp()
        if self.lastTimestamp > timestamp:
            raise Exception()
        if self.lastTimestamp == timestamp:
            self.sequence = (self.sequence + 1) & self.sequenceMask
            if self.sequence == 0:
                timestamp = self.next_timestamp(timestamp)
        else:
            self.sequence = 0
        self.lastTimestamp = timestamp
        return ((timestamp - self.twepoch) << self.timestampLeftShift) | \
               (self.datacenterId << self.datacenterIdShift) | \
               (self.workerId << self.workerIdShift) | \
               self.sequence

    def next_timestamp(self, timestamp, next_timestamp=0):
        while timestamp >= next_timestamp: next_timestamp = self.get_current_timestamp()
        return next_timestamp


id_pool = SnowKey.from_node()



__author__ = 'CzaOrz <https://github.com/CzaOrz>'
from pywss.version import __version__

import socketserver
import json
import ssl

from socket import socket, getdefaulttimeout
from _socket import AF_INET, SOCK_STREAM
from ssl import SSLContext, SSLSocket

from pywss.protocol import WebSocketProtocol
from pywss.middlewares import *
from pywss.public import *
from pywss.route import *

logging.basicConfig(format="%(asctime)s %(funcName)s[lines-%(lineno)d]: %(message)s")
logger = logging.getLogger(__name__)


class WssSSLContext(SSLContext):

    def wrap_socket(self, sock, server_side=False, **kwargs):
        return WssSocket(sock=sock, server_side=server_side, _context=self, **kwargs)


def load_ssl_context(pem, key, version=None):
    ssl_context = WssSSLContext(version or ssl.PROTOCOL_TLSv1)
    ssl_context.load_cert_chain(pem, key)
    return ssl_context


class WsSocket(socket):
    """WsSocket =>
    add new way of ws_recv/ws_send for this class,
    which has been adapted to web-socket-protocol,
    so you can use those as same as recv/send.
    """
    __slots__ = ["_io_refs", "_closed", "__route__", "conn"]

    def __init__(self, family=-1, type=-1, proto=-1, fileno=None):
        if fileno is None:
            if family == -1:
                family = AF_INET
            if type == -1:
                type = SOCK_STREAM
            if proto == -1:
                proto = 0
        super(WsSocket, self).__init__(family, type, proto, fileno)

    def accept(self):
        fd, addr = self._accept()
        sock = WsSocket(self.family, self.type, self.proto, fileno=fd)
        if getdefaulttimeout() is None and self.gettimeout():
            sock.setblocking(True)
        return sock, addr

    def ws_recvall(self, bufsize: int, response=b''):
        chunk = self.recv(bufsize)
        self.setblocking(False)
        while chunk:
            response += chunk
            try:
                chunk = self.recv(bufsize)
            except IOError:
                break
        self.setblocking(True)
        return response

    def ws_recv(self, bufsize: int, flags: int = ...):
        return WebSocketProtocol.decode_msg(self.ws_recvall(bufsize))

    def ws_send(self, data, flags: int = ...):
        self.sendall(WebSocketProtocol.encode_msg(json.dumps(data, ensure_ascii=False).encode('utf-8')))


class WssSocket(SSLSocket, WsSocket):

    def __init__(self, *args, **kwargs):
        if hasattr(self, "_create"):
            kwargs["context"] = kwargs.pop("_context")
            self = self._create(*args, **kwargs)
        else:
            super(WssSocket, self).__init__(*args, **kwargs)


class MyServerTCPServer(socketserver.TCPServer):

    def __init__(self, server_address, RequestHandlerClass, bind_and_activate=True, ssl_context=None, **kwargs):
        socketserver.BaseServer.__init__(self, server_address, RequestHandlerClass)
        self.socket = WsSocket(self.address_family, self.socket_type)
        self.ssl_context = ssl_context
        if bind_and_activate:
            try:
                self.server_bind()
                self.server_activate()
            except:
                self.server_close()
                raise

    def serve_forever(self, poll_interval=0.5):  # todo, amend server for asynchronous
        if mwManager.radio_middleware:  # radio middleware work here
            mwManager.radio_process()
        super(MyServerTCPServer, self).serve_forever(poll_interval)

    def get_request(self):
        request, address = self.socket.accept()
        if self.ssl_context:
            request = self.ssl_context.wrap_socket(request, server_side=True)  # type: WssSocket
        return request, address

    def verify_request(self, request, client_address):
        try:
            path, key = WebSocketProtocol.parsing(request.ws_recvall(1024))
            if path in Route.routes:
                request.send(key)
                request.__route__ = path
                return True
        except:
            pass
        return False


class MyServerThreadingMixIn(socketserver.ThreadingMixIn):
    block_on_close = False


class SocketHandler:

    def __init__(self, request, client_address, server):
        self.request = request
        self.client_address = client_address
        self.func = Route.get(self.request.__route__)
        try:
            self.setup()
            self.handle()
        except:
            pass
        finally:
            self.finish()

    def setup(self):
        self.conn = mwManager.daemon_process(self, self.request)  # type: Connector
        if not self.conn:
            self.conn = Connector(self.request, self.client_address)
        ConnectManager.add_connector(self.conn.name, self.conn.client_address, self.conn)
        logger.info('New Connect {} From {}'.format(self.conn.name, self.conn.client_address))
        if not radio_queue.qsize():
            radio_queue.put(RADIO_START)
        self.request.conn = self.conn

    def handle(self):
        error_count = 0
        try:
            while error_count < PublicConfig.ERROR_COUNT_MAX:
                info = mwManager.process(self.request, self.request.ws_recv(1024), self.func)  # main process
                if info is ERROR_FLAG:
                    error_count += 1
                elif info:
                    self.request.ws_send(info)
                else:
                    error_count += 4
        except:
            pass

    def finish(self):
        try:
            for func in Route.finish:
                func()
        except:
            pass


class Pyws(MyServerThreadingMixIn, MyServerTCPServer):

    def __init__(self,
                 routes_module,
                 address='0.0.0.0', port=8866,
                 RequestHandlerClass=SocketHandler,
                 request_queue_size=5,
                 middleware=None,
                 logging_level=logging.INFO, **kwargs):
        logger.setLevel(logging_level)
        logger.info('Server Start ...')
        logger.info('Server Address is %s:%d' % (address, port))
        server_address = (address, port)
        self.add_routes(routes_module)
        self.add_middleware(middleware)
        self.request_queue_size = request_queue_size
        super(Pyws, self).__init__(server_address, RequestHandlerClass, **kwargs)

    def add_routes(self, module):
        Route.add_routes(module)

    def add_middleware(self, middleware):
        mwManager.auto_add(middleware)

    @staticmethod
    def route(path):
        def wrapper(func):
            Route.add_route(path, func)
            return func

        return wrapper

    @staticmethod
    def after_request(func=None):
        def wrapper(func):
            Route.add_finish(func)
            return func

        if func:
            return wrapper(func)
        return wrapper


class Pywss(Pyws):

    def __init__(self, routes_module, ssl_pem=None, ssl_key=None, ssl_context=None, ssl_version=None, **kwargs):
        if ssl_context:
            pass
        elif ssl_pem and ssl_key:
            ssl_context = load_ssl_context(ssl_pem, ssl_key, ssl_version)
        else:
            raise Exception("you should define ssl file path")
        super(Pywss, self).__init__(routes_module, ssl_context=ssl_context, **kwargs)



import threading
import logging
import time

from collections import deque
from collections.abc import Iterable

from pywss.public import *
from pywss.connector import Connector, ConnectManager
from pywss.public import PublicConfig

logger = logging.getLogger(__name__)


class BaseMiddleware:

    @classmethod
    def process_data(cls): ...

    @classmethod
    def process_input(cls, request, input_msg): return input_msg

    @classmethod
    def process_output(cls, request, output_msg): return output_msg


class RadioMiddleware(BaseMiddleware):

    @classmethod
    def process_data(cls): raise MiddlewareError


class DaemonMiddleware(BaseMiddleware):
    """each conn, just work once"""


class DataMiddleware(BaseMiddleware):
    """each data transform, it will works"""


class MiddlewareManager:
    radio_middleware = []
    daemon_middleware = {
        'process_input': deque(),
        'process_output': deque(),
    }
    data_middleware = {
        'process_input': deque(),
        'process_output': deque(),
    }
    daemon_middleware_count = 0
    data_middleware_count = 0

    @classmethod
    def check_base(cls, middleware):
        if isinstance(middleware, DaemonMiddleware):
            middle_type = 0
        elif isinstance(middleware, DataMiddleware):
            middle_type = 1
        elif isinstance(middleware, RadioMiddleware):
            middle_type = 2
        else:
            raise MiddlewareError
        return middle_type

    @classmethod
    def _add_middleware(cls, attr, middleware):
        getattr(cls, attr)['process_input'].append(getattr(middleware, 'process_input'))
        getattr(cls, attr)['process_output'].appendleft(getattr(middleware, 'process_output'))

    @classmethod
    def add_middleware(cls, middleware=None):
        if isinstance(middleware, type):
            middleware = middleware()
        middleware_type = cls.check_base(middleware)
        if not middleware_type:
            cls.daemon_middleware_count += 1
            cls._add_middleware('daemon_middleware', middleware)
        elif middleware_type is 1:
            cls.data_middleware_count += 1
            cls._add_middleware('data_middleware', middleware)
        elif middleware_type is 2:
            cls.radio_middleware.append(getattr(middleware, 'process_data'))

    @classmethod
    def add_middlewares(cls, middlewares=None):
        if middlewares and isinstance(middlewares, Iterable):
            for middleware in middlewares:
                cls.add_middleware(middleware)

    @classmethod
    def auto_add(cls, middle):
        if middle:
            if isinstance(middle, Iterable):
                cls.add_middlewares(middle)
            else:
                cls.add_middleware(middle)

    @classmethod
    def process(cls, request, data, func=None):
        try:
            for process_input in cls.data_middleware['process_input']:
                data = process_input(request, data)
            data = (func(request, data) if func else data) or PublicConfig.DEFAULT_REPLY
            for process_output in cls.data_middleware['process_output']:
                data = process_output(request, data)
            return data
        except MiddlewareError:
            return ERROR_FLAG

    @classmethod
    def daemon_process(cls, handler, request):
        try:
            if cls.daemon_middleware_count:
                data = request.ws_recv(1024)
                for process_input in cls.daemon_middleware['process_input']:
                    data = process_input(request, data)
                for process_output in cls.daemon_middleware['process_output']:
                    data = process_output(request, data)
                name, clear_level = data if isinstance(data, tuple) else (data, None)
                return Connector(request, handler.client_address, name=name, clear_level=clear_level)
        except:
            raise AuthenticationError

    @classmethod
    def radio_process(cls):
        t = threading.Thread(target=cls._process_radio)
        t.daemon = True
        logger.info('radio is enabled...')
        t.start()

    @classmethod
    def _process_radio(cls):
        """
        each conn, will push `RADIO_START` into `radio_queue` if the queue is empty.
        so this queue size will never exceed two.
        after push, this thread will execute radio middleware as you see.
        :return:
        """
        while True:
            command = radio_queue.get()  # if there is no conn, the radio middleware will blocking
            if command is RADIO_START and ConnectManager.online():
                while ConnectManager.online():
                    logger.info('broadcasting...')
                    for middleware_func in cls.radio_middleware:
                        data = middleware_func()
                        if not data:
                            continue
                        ConnectManager.send_to_all(data)
                    time.sleep(PublicConfig.RADIO_TIME)
                logger.info('radio pause...')


mwManager = MiddlewareManager()



import functools

from importlib import import_module
from pywss.public import InvalidPath


def route(path):
    def wrapper(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        _wrapper.__route__ = path
        return _wrapper

    return wrapper


class Route:
    routes = {}  # all routes for server
    finish = []  # this will work in SocketHandler.finish

    @classmethod
    def add_routes(cls, module_name):
        point = module_name.rfind('.')
        if point == (-1):
            mod = import_module(module_name)
        else:
            mod = getattr(import_module(module_name[:point]), module_name[point + 1:])
        for attr in dir(mod):
            if attr.startswith('_'):
                continue
            func = getattr(mod, attr)
            path = getattr(func, '__route__', None)
            if path and callable(func):
                cls.add_route(path, func)

    @classmethod
    def add_route(cls, path, func):
        cls.routes.setdefault(path, func)

    @classmethod
    def add_finish(cls, func):
        cls.finish.append(func)

    @classmethod
    def get(cls, path):
        func = cls.routes.get(path)
        if not func:
            raise InvalidPath
        return func



import logging

from weakref import finalize
from collections import defaultdict
from threading import current_thread

from pywss.snow_key import id_pool

logger = logging.getLogger(__name__)


class Connector:
    def __init__(self, request, client_address, name=None, clear_level=None):
        self.request = request
        self.client_address = ':'.join([str(i) for i in client_address])
        if name:
            self.name = name
            self.clear_level = 0 if clear_level is None else clear_level
        else:
            self.name = str(id_pool.next_id())
            self.clear_level = 1 if clear_level is None else clear_level
        self.send_to_connector = send_to_connector
        self.send_to_all = send_to_all


class ConnectManager:
    connectors = defaultdict(dict)

    @classmethod
    def online(cls):
        return len(cls.connectors)

    @classmethod
    def exists(cls, name):
        return name in cls.connectors and cls.connectors[name].values()

    @classmethod
    def add_connector(cls, name, key, value):

        def remove_connectors(name=name, key=key, clear_level=value.clear_level):
            try:
                if not key:
                    cls.connectors.pop(name)
                if key in cls.connectors[name]:
                    if clear_level:
                        cls.connectors.pop(name)
                    else:
                        cls.connectors[name].pop(key)
            except:
                pass
            finally:
                logger.warning('{} Connect Close'.format(key))

        finalize(current_thread(), remove_connectors)
        cls.connectors[name][key] = value

    @classmethod
    def send_to_connector(cls, name, msg):
        try:
            for connector in cls.connectors[name].values():
                try:
                    connector.request.ws_send(msg)
                except OSError:
                    continue
            logger.info('send msg success')
            return True
        except:
            return False

    @classmethod
    def send_to_all(cls, msg):
        try:
            for connector in cls.next_user():
                try:
                    connector.request.ws_send(msg)
                except OSError:
                    continue
            logger.info('send msg success')
            return True
        except:
            return False

    @classmethod
    def next_user(cls):
        return [connector for connectors in cls.connectors.values()
                for connector in connectors.values()]


send_to_connector = ConnectManager.send_to_connector
send_to_all = ConnectManager.send_to_all



import os
import re
import codecs

from setuptools import setup, find_packages

"""pip install twine
1、python setup check
2、python setup sdist
3、twine upload dist/__packages__-__version__.tar.gz
"""


def read(*parts):
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, *parts), 'r', encoding='utf-8').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='pywss',
    version=find_version('pywss', 'version.py'),
    description="This is a web-socket-server by python",
    long_description="see https://github.com/CzaOrz/Pywss",
    author='czaOrz',
    author_email='chenziangsg@163.com',
    url='https://github.com/CzaOrz/Pywss',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)
